# -*- coding: utf-8 -*-

import smtplib
import logging
import traceback

from django.utils import timezone as tz
from django.conf import settings
from django.contrib.sites.models import Site

from crma.models import EmailScheduler, ST_SENT, ST_ERROR
from crma.mails import MailBuilder, BadEmailAddress
from crma.signals import email_sent, emails_sent
from django.core import mail


logger = logging.getLogger(__name__)


class ConnectionManager(object):
    def __init__(self, connection, messages):
        self.connection = connection
        self.messages = messages
        self.index = 0
        self._finished = False

    def __call__(self):
        mails = []
        while not self._finished:
            try:
                self.connection.open()
                mails.extend(self.send_messages())
            except smtplib.SMTPServerDisconnected:
                continue
            finally:
                self.connection.close()
        return mails

    def get_messages(self):
        return self.messages[self.index:]

    def send_messages(self):
        for index, message in enumerate(self.get_messages(), start=self.index):
            try:
                message.send(self.connection)
            except smtplib.SMTPServerDisconnected:
                logger.warning('Disconnected after %s mails', index - self.index)
                self.index = index
                raise
            except Exception as e:
                logger.warning('Cannot send %s: %s(%s)', message, type(e).__name__, e)
            else:
                yield message
        self.index = index + 1
        self._finished = self.index >= len(self.messages)


class ConnectionManagerBunch(ConnectionManager):
    def __init__(self, connection, messages, max):
        super(ConnectionManagerBunch, self).__init__(connection, messages)
        self.max = max

    def get_messages(self):
        return super(ConnectionManagerBunch, self).get_messages()[:self.max]


class SendableMessage(object):
    def __init__(self, email_scheduler):
        self.email_scheduler = email_scheduler
        self.message = None

    def __repr__(self):
        return 'M<{!r}>'.format(self.email_scheduler)

    def _set_error(self, trace):
        self.email_scheduler.status = ST_ERROR
        self.email_scheduler.trace_error = traceback.format_exc()
        self.email_scheduler.save()

    def prepare(self):
        mb = MailBuilder(self.email_scheduler, Site.objects.get_current())
        try:
            message = mb.build_email()
        except BadEmailAddress as e:
            logging.debug('Skip mail to %s', e)
            self._set_error(e)
        except Exception as e:
            logger.exception('Error preparing %s', self)
            self._set_error(e)
            raise
        self.message = message
        return message

    def send(self, connection):
        assert self.message, 'Message has not been prepared'
        try:
            connection.send_messages([self.message])
        except smtplib.SMTPServerDisconnected:
            raise
        except smtplib.SMTPRecipientsRefused as e:
            logger.error('Error sending %s', self)
            self._set_error(e)
            raise
        except smtplib.SMTPResponseException as e:
            logger.error('Error sending %s', self)
            if e.smtp_code >= 500:
                self._set_error(e)
            raise
        except Exception as e:
            logger.exception('Error sending %s', self)
            self._set_error(e)
            raise
        else:
            self._set_sent()
        email_sent.send(sender=EmailScheduler, data={
            'id': self.email_scheduler.id,
            'email_scheduler': self.email_scheduler,
            'message': self.message,
        })

    def _set_sent(self):
        self.email_scheduler.status = ST_SENT
        self.email_scheduler.sent_time = tz.now()
        self.email_scheduler.save()


class Mailer(object):
    def __init__(self, date_fn=tz.now):
        self.date_fn = date_fn

    def __call__(self):
        # Emails to send
        now = self.date_fn()
        email_schedulers = EmailScheduler.objects.sendable(now)
        prepared_emails = self.prepare_emails(email_schedulers)
        if not prepared_emails:
            logger.debug('Nothing to do')
            return

        logger.info('%s - %s prepared emails', now, len(prepared_emails))
        connection = mail.get_connection(getattr(settings, 'CRMA_EMAIL_BACKEND', None))
        max_mail_by_connection = getattr(settings, 'CRMA_MAX_MAILS_BY_CONNECTION', 0)
        if max_mail_by_connection > 0:
            sender = ConnectionManagerBunch(connection, prepared_emails, max_mail_by_connection)
        else:
            sender = ConnectionManager(connection, prepared_emails)

        sent_mails = list(sender())
        logger.info('%s - %s sent emails', now, len(sent_mails))

        emails_sent.send(sender=EmailScheduler,
                         prepared_emails=prepared_emails,
                         emails=sent_mails,
                         )

    def prepare_emails(self, email_schedulers):
        if email_schedulers.count() == 0:
            return []

        messages = []
        for email_scheduler in email_schedulers.iterator():
            sm = SendableMessage(email_scheduler)
            try:
                sm.prepare()
            except Exception as e:
                logger.warning('Cannot prepare %s: %s', email_scheduler, e)
            else:
                messages.append(sm)
        return messages
